/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::MemoryPlotter
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "MemoryPlotter.h"

#include <ArmarXCore/core/time/Frequency.h>
#include <ArmarXCore/core/time/Metronome.h>

#include <RobotAPI/libraries/armem/core/error/mns.h>
#include <RobotAPI/libraries/aron/core/data/variant/Variant.h>
#include <RobotAPI/libraries/aron/core/data/variant/primitive/All.h>
#include <RobotAPI/libraries/aron/core/data/visitor/variant/VariantVisitor.h>

namespace armarx
{


    armarx::PropertyDefinitionsPtr
    MemoryPlotter::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs =
            new ComponentPropertyDefinitions(getConfigIdentifier());


        return defs;
    }

    MemoryPlotter::MemoryPlotter()
    {
    }

    std::string
    MemoryPlotter::getDefaultName() const
    {
        return "MemoryPlotter";
    }

    void
    MemoryPlotter::onInitComponent()
    {
        DebugObserverComponentPluginUser::setDebugObserverBatchModeEnabled(true);

        armem::MemoryID robotEntityId{"RobotState", "Proprioception", "Armar7", "Armar7"};

        std::vector<std::string> jointNames{
            "Neck_1_Yaw",
            "Neck_2_Hemisphere_A",
            "Neck_3_Hemisphere_B",
        };

        for (const std::string& jointName : jointNames)
        {
            std::vector<std::string> attributes{
                "position",
                "velocity",
            };
            for (const std::string& attribute : attributes)
            {
                properties.plottedValues.push_back(MemoryValueID{
                    .entityID = robotEntityId,
                    .aronPath = {{"joints", attribute, jointName}},
                });
            }
        }
    }

    void
    MemoryPlotter::onConnectComponent()
    {
        createRemoteGuiTab();
        RemoteGui_startRunningTask();

        task = new RunningTask<MemoryPlotter>(this, &MemoryPlotter::runLoop);
        task->start();
    }

    void
    MemoryPlotter::onDisconnectComponent()
    {
        task->stop();
    }

    void
    MemoryPlotter::onExitComponent()
    {
    }

    void
    MemoryPlotter::runLoop()
    {
        Frequency frequency = Frequency::Hertz(30);
        Metronome metronome(frequency);
        while (not task->isStopped())
        {
            loopOnce();
            metronome.waitForNextTick();
        }
    }

    class Visitor : public aron::data::ConstVariantVisitor
    {
    public:
        Visitor(armarx::plugins::DebugObserverComponentPlugin& debugObserver) :
            debugObserver{debugObserver}
        {
        }

        // void visitAronVariant(const data::DictPtr&) override;
        // void visitAronVariant(const data::ListPtr&) override;
        // void visitAronVariant(const data::NDArrayPtr&) override;
        void
        visitAronVariant(const aron::data::IntPtr& v) override
        {
            setDatafield(v->getValue());
        }

        void
        visitAronVariant(const aron::data::LongPtr& v) override
        {
            setDatafield(v->getValue());
        }

        void
        visitAronVariant(const aron::data::FloatPtr& v) override
        {
            setDatafield(v->getValue());
        }

        void
        visitAronVariant(const aron::data::DoublePtr& v) override
        {
            setDatafield(v->getValue());
        }

        void
        visitAronVariant(const aron::data::BoolPtr& v) override
        {
            setDatafield(v->getValue());
        }

        void
        visitAronVariant(const aron::data::StringPtr& v) override
        {
            setDatafield(v->getValue());
        }

        template <class ValueT>
        void
        setDatafield(const ValueT& value)
        {
            debugObserver.setDebugObserverDatafield(channelName, datafieldName, value);
            ++count;
        }

        armarx::plugins::DebugObserverComponentPlugin& debugObserver;
        std::string channelName;
        std::string datafieldName;

        int count = 0;
    };

    void
    MemoryPlotter::loopOnce()
    {
        Visitor visitor(getDebugObserverComponentPlugin());

        std::stringstream log;
        for (const MemoryValueID& valueId : properties.plottedValues)
        {
            armem::client::Reader* reader = getReader(valueId.entityID, log);
            if (reader == nullptr)
            {
                // Log is written in getReader.
                continue;
            }

            std::optional<armem::wm::EntitySnapshot> snapshot =
                reader->getLatestSnapshotIn(valueId.entityID);

            if (not snapshot.has_value())
            {
                log << "\nDid not find snapshot in entity " << valueId.entityID << ".";
                continue;
            }

            aron::data::DictPtr data = snapshot->getInstance(0).data();
            aron::data::VariantPtr valueVariant = data->navigateAbsolute(valueId.aronPath);
            if (not valueVariant)
            {
                log << "\nDid not find " << valueId.aronPath.toString() << " in entity snapshot "
                    << snapshot->id() << ".";
                continue;
            }


            auto makeDatafieldName = [](const aron::Path& path) -> std::string
            {
                // The first element is always "ARON->", which we can discard for readability.
                std::string str = path.toString();
                str =
                    simox::alg::remove_prefix(str, path.getRootIdentifier() + path.getDelimeter());
                str = simox::alg::replace_all(str, path.getDelimeter(), ">");
                return str;
            };

            visitor.channelName = simox::alg::replace_all(valueId.entityID.str(), "/", ">");
            visitor.datafieldName = makeDatafieldName(valueId.aronPath);

            aron::data::visit(visitor, valueVariant);
        }

        ARMARX_INFO << "Sending debug observer batch with " << visitor.count << " values ...";
        sendDebugObserverBatch();

        if (not log.str().empty())
        {
            ARMARX_INFO << "Encountered issues while sending memory values to the debug observer "
                           "for plotting: "
                        << log.str();
        }
    }

    armem::client::Reader*
    MemoryPlotter::getReader(const armem::MemoryID& memoryID, std::stringstream& log)
    {
        armem::MemoryID key = memoryID.getProviderSegmentID();

        auto it = memoryReaders.find(key);
        if (it != memoryReaders.end())
        {
            return &it->second;
        }
        else
        {
            armem::client::Reader reader;
            try
            {
                reader = memoryNameSystem().getReader(key);
            }
            catch (armem::error::CouldNotResolveMemoryServer& e)
            {
                log << "\n" << e.what();
                return nullptr;
            }

            auto [it, _] = memoryReaders.emplace(key, reader);
            return &it->second;
        }
    }

    void
    MemoryPlotter::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        if (tab.queryResult)
        {
        }

        VBoxLayout root = {tab.queryResultGroup, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }

    void
    MemoryPlotter::RemoteGui_update()
    {
        if (tab.rebuild.exchange(false))
        {
            createRemoteGuiTab();
        }
    }

} // namespace armarx
