/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::MemoryPlotter
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once


#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/interface/observers/ObserverInterface.h>
#include <ArmarXCore/libraries/ArmarXCoreComponentPlugins/DebugObserverComponentPlugin.h>
#include <ArmarXCore/util/tasks.h>

#include <ArmarXGui/libraries/ArmarXGuiComponentPlugins/LightweightRemoteGuiComponentPlugin.h>

#include <RobotAPI/interface/armem/mns/MemoryNameSystemInterface.h>
#include <RobotAPI/libraries/armem/client/Reader.h>
#include <RobotAPI/libraries/armem/client/plugins/PluginUser.h>
#include <RobotAPI/libraries/armem/core/wm/memory_definitions.h>

namespace armarx
{


    struct MemoryValueID
    {
        armem::MemoryID entityID;
        aron::Path aronPath;
    };

    /**
     * @defgroup Component-MemoryPlotter MemoryPlotter
     * @ingroup RobotAPI-Components
     *
     * Transfers data from the memory system to the DebugObserver, allowing to visualize them in
     * the LivePlotter.
     *
     * @class MemoryPlotter
     * @ingroup Component-MemoryPlotter
     * @brief Implementation of \ref Component-MemoryPlotter.
     */
    class MemoryPlotter :
        virtual public armarx::Component,
        virtual public armarx::armem::ClientPluginUser,
        virtual public armarx::DebugObserverComponentPluginUser,
        virtual public armarx::LightweightRemoteGuiComponentPluginUser
    {
    public:
        MemoryPlotter();

        /// @see armarx::ManagedIceObject::getDefaultName()
        std::string getDefaultName() const override;

        // LightweightRemoteGuiComponentPluginUser interface
    public:
        void createRemoteGuiTab();
        void RemoteGui_update() override;


    protected:
        armarx::PropertyDefinitionsPtr createPropertyDefinitions() override;

        void onInitComponent() override;
        void onConnectComponent() override;
        void onDisconnectComponent() override;
        void onExitComponent() override;

        void runLoop();
        void loopOnce();

        armem::client::Reader* getReader(const armem::MemoryID& memoryID, std::stringstream& log);


    private:
        struct Properties
        {
            std::vector<MemoryValueID> plottedValues;
        };

        Properties properties;

        std::map<armem::MemoryID, armem::client::Reader> memoryReaders;

        armarx::RunningTask<MemoryPlotter>::pointer_type task;

        struct RemoteGuiTab : RemoteGui::Client::Tab
        {
            std::atomic_bool rebuild = false;

            std::optional<armem::wm::Memory> queryResult;
            RemoteGui::Client::GroupBox queryResultGroup;
        };

        RemoteGuiTab tab;
    };
} // namespace armarx
